#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

#include "itkImageRegistrationMethodv4.h"
#include "itkMattesMutualInformationImageToImageMetricv4.h"

#include "itkAffineTransform.h"
#include "itkANTSNeighborhoodCorrelationImageToImageMetricv4.h"
#include "itkBSplineTransform.h"
#include "itkBSplineTransformParametersAdaptor.h"
#include <itkConjugateGradientLineSearchOptimizerv4.h>
#include "itkMattesMutualInformationImageToImageMetricv4.h"


template <class ImageType>
typename ImageType::Pointer 
ReadImage(const std::string& filename);


template<class TFilter, class TOptimizer>
class CommandIterationUpdate : public itk::Command
{
public:
  typedef CommandIterationUpdate   Self;
  typedef itk::Command             Superclass;
  typedef itk::SmartPointer<Self>  Pointer;
  itkNewMacro( Self );

protected:
  CommandIterationUpdate() {};

public:

  void Execute(itk::Object *caller, const itk::EventObject & event){
    Execute( (const itk::Object *) caller, event);
  }

  void Execute(const itk::Object * object, const itk::EventObject & event){
    const TFilter * filter = dynamic_cast< const TFilter * >( object );
    if( typeid( event ) != typeid( itk::IterationEvent ) )
    { return; }

    unsigned int currentLevel = filter->GetCurrentLevel();
    typename TFilter::ShrinkFactorsPerDimensionContainerType shrinkFactors = filter->GetShrinkFactorsPerDimension( currentLevel );
    typename TFilter::SmoothingSigmasArrayType smoothingSigmas = filter->GetSmoothingSigmasPerLevel();
    typename TFilter::TransformParametersAdaptorsContainerType adaptors = filter->GetTransformParametersAdaptorsPerLevel();

    typename itk::ObjectToObjectOptimizerBase::Pointer optimizerBase = (const_cast<TFilter*>(filter))->GetModifiableOptimizer();
    typename TOptimizer::Pointer optimizer = dynamic_cast<TOptimizer *>(optimizerBase.GetPointer());
    if( !optimizer ){
      itkGenericExceptionMacro( "Error dynamic_cast failed" );
    }
    typename TOptimizer::DerivativeType gradient = optimizer->GetGradient();

    /* orig
    std::cout << "  Current level = " << currentLevel << std::endl;
    std::cout << "    shrink factor = " << shrinkFactors[currentLevel] << std::endl;
    std::cout << "    smoothing sigma = " << smoothingSigmas[currentLevel] << std::endl;
    std::cout << "    required fixed parameters = " << adaptors[currentLevel]->GetRequiredFixedParameters() << std::endl;
    */

    //debug:
    std::cout << "  CL Current level:           " << currentLevel << std::endl;
    std::cout << "   SF Shrink factor:          " << shrinkFactors << std::endl;
    std::cout << "   SS Smoothing sigma:        " << smoothingSigmas[currentLevel] << std::endl;
    std::cout << "   RFP Required fixed params: " << adaptors[currentLevel]->GetRequiredFixedParameters() << std::endl;
    std::cout << "   LR Final learning rate:    " << optimizer->GetLearningRate() << std::endl;
    std::cout << "   FM Final metric value:     " << optimizer->GetCurrentMetricValue() << std::endl;
    std::cout << "   SC Optimizer scales:       " << optimizer->GetScales() << std::endl;
    std::cout << "   FG Final metric gradient (sample of values): ";
    if( gradient.GetSize() < 10 ){
      std::cout << gradient;
    }else{
      for( itk::SizeValueType i = 0; i < gradient.GetSize(); i += (gradient.GetSize() / 16) ){
        std::cout << gradient[i] << " ";
      }// for
    }
    std::cout << std::endl;
  } //Execute
};
//-----------------------------------------------------------------------------

int main( int argc, char *argv[] )
{

  std::string fixed_image_file = "..\\data\\RatLungSlice1.mha";
  std::string moving_image_file = "..\\data\\RatLungSlice2.mha";
  
  std::string deformed_image_file = "..\\output\\RatLungSlice2_mi_ffd_gd.mha";
  unsigned int numberOfDeformableIterations = 200;
  bool use_conjugate_gls_optimizer = true;
  
  const unsigned int VImageDimension = 2;
  typedef double                PixelType;
  typedef double								RealType;
  
  typedef itk::Image<PixelType, VImageDimension> FixedImageType;
  typedef itk::Image<PixelType, VImageDimension> MovingImageType;

  
  // read image
  FixedImageType::Pointer fixedImage = ReadImage<FixedImageType>(fixed_image_file);
  MovingImageType::Pointer movingImage = ReadImage<MovingImageType>(moving_image_file);

  //
  // b-spline deformable transform with Mattes MI metric
  typedef itk::MattesMutualInformationImageToImageMetricv4<FixedImageType, MovingImageType> MutualInformationMetricType;
  MutualInformationMetricType::Pointer mutualInformationMetric = MutualInformationMetricType::New();
  mutualInformationMetric = mutualInformationMetric;
  mutualInformationMetric->SetNumberOfHistogramBins( 32 );
  mutualInformationMetric->SetUseMovingImageGradientFilter( false );
  mutualInformationMetric->SetUseFixedImageGradientFilter( false );
  mutualInformationMetric->SetUseFixedSampledPointSet( false );
          
  typedef itk::RegistrationParameterScalesFromPhysicalShift<MutualInformationMetricType> ScalesEstimatorType;
  ScalesEstimatorType::Pointer scalesEstimator = ScalesEstimatorType::New();
  scalesEstimator->SetMetric( mutualInformationMetric );
  scalesEstimator->SetTransformForward( true );

  typedef itk::GradientDescentOptimizerv4           Optimizerv4Type;
  Optimizerv4Type::Pointer optimizer = 0;

  if(use_conjugate_gls_optimizer){
    typedef itk::ConjugateGradientLineSearchOptimizerv4  ActualOptimizerv4Type;
    ActualOptimizerv4Type::Pointer actual_optimizer = ActualOptimizerv4Type::New();
    actual_optimizer->SetLowerLimit( 0 );
    actual_optimizer->SetUpperLimit( 2 );
    actual_optimizer->SetEpsilon( 0.2 );
    //actual_optimizer->SetMaximumLineSearchIterations( 20 );
    actual_optimizer->SetLearningRate(  1.0  );
    actual_optimizer->SetMaximumStepSizeInPhysicalUnits(  1.0  );
    actual_optimizer->SetNumberOfIterations(numberOfDeformableIterations);
    actual_optimizer->SetScalesEstimator( scalesEstimator );
    actual_optimizer->SetMinimumConvergenceValue( 1e-7 ); // default 1e-8  --> 1e-6
    actual_optimizer->SetConvergenceWindowSize( 50 ); // default 50 --> 10
    actual_optimizer->SetDoEstimateLearningRateAtEachIteration( true );
    actual_optimizer->SetDoEstimateLearningRateOnce( false);  
    optimizer = actual_optimizer;
  }else{
    typedef Optimizerv4Type ActualOptimizerv4Type;
    ActualOptimizerv4Type::Pointer actual_optimizer = ActualOptimizerv4Type::New();
    actual_optimizer->SetLearningRate( 1.0 );
    actual_optimizer->SetNumberOfIterations(numberOfDeformableIterations);
    actual_optimizer->SetScalesEstimator( scalesEstimator );
    actual_optimizer->SetDoEstimateLearningRateOnce( false ); //true by default
    actual_optimizer->SetDoEstimateLearningRateAtEachIteration( true );
    optimizer = actual_optimizer;
  }

  const unsigned int numberOfLevels = 3;
  const unsigned int SplineOrder = 3;
  typedef itk::BSplineTransform<RealType, VImageDimension, SplineOrder> BSplineTransformType;

  typedef itk::ImageRegistrationMethodv4<FixedImageType, MovingImageType, BSplineTransformType> BSplineRegistrationType;
  BSplineRegistrationType::Pointer bsplineRegistration = BSplineRegistrationType::New();

  // Shrink the virtual domain by specified factors for each level. 
  BSplineRegistrationType::ShrinkFactorsArrayType shrinkFactorsPerLevel;
  shrinkFactorsPerLevel.SetSize( 3 );
  shrinkFactorsPerLevel[0] = 3;
  shrinkFactorsPerLevel[1] = 2;
  shrinkFactorsPerLevel[2] = 1;

  // Smooth by specified gaussian sigmas for each level.  These values are specified in
  // physical units.
  BSplineRegistrationType::SmoothingSigmasArrayType smoothingSigmasPerLevel;
  smoothingSigmasPerLevel.SetSize( 3 );
  smoothingSigmasPerLevel[0] = 2;
  smoothingSigmasPerLevel[1] = 1;
  smoothingSigmasPerLevel[2] = 1;

  BSplineTransformType::Pointer outputBSplineTransform = const_cast<BSplineTransformType *>( bsplineRegistration->GetOutput()->Get() );

  BSplineTransformType::PhysicalDimensionsType physicalDimensions;
  BSplineTransformType::MeshSizeType meshSize;
  for( unsigned int d = 0; d < VImageDimension; d++ ){
    physicalDimensions[d] = fixedImage->GetSpacing()[d] * static_cast<RealType>( fixedImage->GetLargestPossibleRegion().GetSize()[d] - 1 );
    meshSize[d] = 5;
  }

  // Create the transform adaptors

  typedef itk::BSplineTransformParametersAdaptor<BSplineTransformType> BSplineTransformAdaptorType;
  BSplineRegistrationType::TransformParametersAdaptorsContainerType adaptors;
  // Create the transform adaptors specific to B-splines
  for(unsigned int level = 0; level < numberOfLevels; level++ ){
    typedef itk::ShrinkImageFilter<FixedImageType, FixedImageType> ShrinkFilterType;
    ShrinkFilterType::Pointer shrinkFilter = ShrinkFilterType::New();
    shrinkFilter->SetShrinkFactors( shrinkFactorsPerLevel[level] );
    shrinkFilter->SetInput( fixedImage );
    shrinkFilter->Update();

    // A good heuristic is to double the b-spline mesh resolution at each level
    BSplineTransformType::MeshSizeType requiredMeshSize;
    for( unsigned int d = 0; d < VImageDimension; d++ ){
      requiredMeshSize[d] = meshSize[d] << level;
    }
    typedef itk::BSplineTransformParametersAdaptor<BSplineTransformType> BSplineAdaptorType;
    BSplineAdaptorType::Pointer bsplineAdaptor = BSplineAdaptorType::New();
    bsplineAdaptor->SetTransform( outputBSplineTransform );
    bsplineAdaptor->SetRequiredTransformDomainMeshSize( requiredMeshSize );
    bsplineAdaptor->SetRequiredTransformDomainOrigin( shrinkFilter->GetOutput()->GetOrigin() );
    bsplineAdaptor->SetRequiredTransformDomainDirection( shrinkFilter->GetOutput()->GetDirection() );
    bsplineAdaptor->SetRequiredTransformDomainPhysicalDimensions( physicalDimensions );

    adaptors.push_back( bsplineAdaptor.GetPointer() );
  }// for level

  bsplineRegistration->SetFixedImage( 0, fixedImage );
  bsplineRegistration->SetMovingImage( 0, movingImage );
  bsplineRegistration->SetMetric( mutualInformationMetric );
  bsplineRegistration->SetNumberOfLevels( numberOfLevels );
  bsplineRegistration->SetSmoothingSigmasPerLevel( smoothingSigmasPerLevel );
  bsplineRegistration->SetShrinkFactorsPerLevel( shrinkFactorsPerLevel );
  bsplineRegistration->SetOptimizer( optimizer );
  bsplineRegistration->SetTransformParametersAdaptorsPerLevel( adaptors );

  outputBSplineTransform->SetTransformDomainOrigin( fixedImage->GetOrigin() );
  outputBSplineTransform->SetTransformDomainPhysicalDimensions( physicalDimensions );
  outputBSplineTransform->SetTransformDomainMeshSize( meshSize );
  outputBSplineTransform->SetTransformDomainDirection( fixedImage->GetDirection() );
  outputBSplineTransform->SetIdentity();

  typedef CommandIterationUpdate<BSplineRegistrationType, Optimizerv4Type> BSplineRegistrationCommandType;
  BSplineRegistrationCommandType::Pointer bsplineObserver = BSplineRegistrationCommandType::New();
  bsplineRegistration->AddObserver( itk::IterationEvent(), bsplineObserver );

  try{
    std::cout << "BSpline. txf - bspline update" << std::endl;
    bsplineRegistration->Update();
  }catch( itk::ExceptionObject &e ){
    std::cerr << "Exception caught: " << e << std::endl;
    return EXIT_FAILURE;
  }
  BSplineTransformType::Pointer output_tfm = const_cast<BSplineTransformType *>( bsplineRegistration->GetOutput()->Get());

  std::cout << "After displacement registration: " << std::endl
            << "Last LearningRate: " << optimizer->GetLearningRate() << std::endl
            << "Use FltPtCorrex: " << mutualInformationMetric->GetUseFloatingPointCorrection() << std::endl
            << "FltPtCorrexRes: " << mutualInformationMetric->GetFloatingPointCorrectionResolution() << std::endl
            << "Number of threads used: metric: " << mutualInformationMetric->GetNumberOfThreadsUsed()
            << "Number of threads used: metric: " << mutualInformationMetric->GetNumberOfThreadsUsed()
            << " optimizer: " << bsplineRegistration->GetOptimizer()->GetNumberOfThreads() << std::endl;

  typedef itk::ResampleImageFilter<MovingImageType, FixedImageType> ResampleFilterType;
  ResampleFilterType::Pointer resampler = ResampleFilterType::New();
  resampler->SetTransform( output_tfm );
  resampler->SetInput( movingImage );
  resampler->SetSize( fixedImage->GetLargestPossibleRegion().GetSize() );
  resampler->SetOutputOrigin(  fixedImage->GetOrigin() );
  resampler->SetOutputSpacing( fixedImage->GetSpacing() );
  resampler->SetOutputDirection( fixedImage->GetDirection() );
  resampler->SetDefaultPixelValue( 0 );
  resampler->Update();

  typedef itk::ImageFileWriter<FixedImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName(deformed_image_file);
  writer->SetInput( resampler->GetOutput() );
  writer->Update();

  return EXIT_SUCCESS;
}

//-----------------------------------------------------------------------------
template <class ImageType>
typename ImageType::Pointer ReadImage(const std::string& filename)
{
  typedef itk::ImageFileReader<ImageType>                   ImageReaderType;

  typename ImageReaderType::Pointer image_reader = ImageReaderType::New();
  typename ImageType::Pointer image = 0;

  image_reader->SetFileName(filename.c_str());
  try{
    image_reader->Update();
    image = image_reader->GetOutput();
    image->DisconnectPipeline();
  }catch(itk::ExceptionObject& err){ 
    std::cout << "Could not read one of the input images." << std::endl;
    std::cout << err << std::endl;
    exit(EXIT_FAILURE);
  }
  return image;
}
//-----------------------------------------------------------------------------
